from . import Method
from ..exceptions import RequiredArgument
from ..settings import match, merge, pprint, Settings
from collections.abc import Mapping, ValuesView
from json import dumps as tojson
from os.path import dirname
from pathlib import Path
import errno

class Update(Method):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key + '/update')

  def __call__(self, args:Mapping=None, defaults:bool=False, dry_run:bool=False, overwrite:bool=False):
    resource = self.owner
    context = str(self.context) + '()'
    props = resource.properties

    # Deep copy the arguments and prepare all auxiliary arguments
    args_ = props(args, defaults=defaults)

    # Use only primary key arguments to resolve the source file name
    pk_ = {}
    required = set()
    pk = props.primarykey(args_, pk_, required)

    if required:
      pprint({context + '.args': args_})
      raise RequiredArgument(context, ', '.join(required))

    merge(pk_, pk)

    try:
      # Load the current Terraform, if exists
      source = Path(resource.format('source', pk))
    except Exception as e:
      pprint({context + '.primary_key': pk})
      raise e

    try:
      terraform = resource.load(source)
      init = False
    except FileNotFoundError:
      terraform = Settings()
      init = True

    try:
      # Resolve the item's address within Terraform
      address = resource.format('address', args_)
    except Exception as e:
      pprint({context + '.arguments': args_})
      raise e

    # Control variable to know if it is creating or updating the resource
    settings = None

    if terraform:
      items = terraform.get(address)

      if items is None and address != resource.address:
        address_ = dirname(address)

        if address_ != '':
          items = terraform.get(address_)

          if isinstance(items, Mapping):
            items = items.values()

      if isinstance(items, (list, ValuesView)):
        # Using primary key to match items
        settings_ = props.tosettings(pk_, defaults=False)

        if settings_:
          for _ in items:
            # Exact match, no wildcards
            if match(_, settings_, literally=True, default=True):
              settings = _
              break
      else:
        settings = items

      # Discard old settings if `overwrite` option is enabled
      if isinstance(settings, Mapping) and overwrite:
        settings.clear()

    if settings:
      if defaults:
        raise FileExistsError(errno.EEXIST, '{}: Object already exists'.format(resource.name), tojson(pk))
      exists = True
    else:
      exists = False

    # Generate the template
    template = resource.template(args_)
    terraform.merge((template.copy() if dry_run else template), extend=True, clone=False)

    try:
      if exists:
        settings_ = props.tosettings(args_, defaults=False)
        merge(settings, settings_, extend=True, clone=False)
        action = 'Updated'
        created = False
      else:
        if not defaults:
          args_ = props(args)

        try:
          # Convert arguments to settings and validates the request as well
          settings_ = props.tosettings(args_)
        except ValueError as e:
          raise RequiredArgument(resource.name, *e.args)

        if settings is None:
          settings = settings_
          terraform.merge(Settings().update({address: settings}), extend=True, clone=False)
          action = 'Created'
        else:
          merge(settings, settings_, clone=False)
          action = 'Overwritten'

        created = True
    except Exception as e:
      pprint({context + '.arguments': args_})
      raise e

    conflicts_with = resource.conflicts_with

    if conflicts_with:
      if isinstance(conflicts_with, str):
        conflicts_with = [conflicts_with]

      for key in conflicts_with:
        terraform.pop(key)

    if dry_run:
      settings = settings_

    try:
      resource.beforesave(settings)
    except Exception as e:
      pprint({context + '.settings': settings})
      raise e

    if not dry_run:
      try:
          source_dir = source.parent
          source_dir.mkdir(parents=True, exist_ok=False)
          print(context + ': mkdir', str(source_dir))
      except FileExistsError:
        pass

    if terraform:
      if dry_run:
        template.merge(Settings().update({address: settings}), extend=True, clone=False).print(explicit_start=True, sort_keys=True)
        print('---')
      else:
        # Save the new Terraform to the file
        resource.dump(str(source), terraform, sort_keys=True)

      print(context + ':', action, source, address)

      heritage = props.heritage(args_)
      _ = merge({}, args, heritage)

      if init:
        resource.trigger('init', _, overwrite=overwrite, dry_run=dry_run)

      resource.trigger('change', _, overwrite=overwrite, dry_run=dry_run)

      if created:
        resource.trigger('create', _, overwrite=overwrite, dry_run=dry_run)
      else:
        resource.trigger('update', _, overwrite=overwrite, dry_run=dry_run)

    # --------------------------------------------------------------------------

    # Make sure the resource is a module
    if resource.module.file:
      try:
        filename = Path(resource.module.format('file', pk))
      except (Exception) as e:
        pprint({context + '.primary_key': pk})
        raise e

      try:
        address, settings_ = resource.module(args_)
      except (Exception) as e:
        pprint({context + '.arguments': args_})
        raise e

      # Load the current Terraform, if exists
      try:
        terraform = resource.module.load(str(filename))
      except FileNotFoundError:
        terraform = Settings()

      settings = terraform.get(address)

      if settings:
        if overwrite:
          settings.clear()
          action = 'Overwritten'
        else:
          action = 'Updated'
      else:
        action = 'Created'

      terraform_ = Settings().update({address: settings_})

      if dry_run:
        # Only print the object that would be saved, without saving it.
        terraform_.print(explicit_start=True, sort_keys=True)
        print('---')
      else:
        try:
          filename.parent.mkdir(parents=True, exist_ok=False)
          print(context + ': mkdir', str(filename.parent))
        except FileExistsError:
          pass

        resource.module.dump(str(filename), terraform.merge(terraform_, clone=False), sort_keys=True)

      print(context + ':', action, str(filename), address)

    return args_

class Create(Update):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key)
    self.key = self.key.with_name('create')

  def __call__(self, args:Mapping=None, **opts):
    opts['defaults'] = True
    return super().__call__(args, **opts)



from .settings import get, Settings, Descriptor
from .template import Template
from os.path import dirname, join as joinpath
from pathlib import Path, PurePosixPath
from collections.abc import Mapping
from json import dumps as tojson
import hashlib

class Module(Settings):
  args = Descriptor('args', [])
  file = Descriptor('file')
  name = Descriptor('name')
  source = Descriptor('source')
  template = Descriptor('template')

  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(get(cfg, key))
    self.owner = owner
    self.key = PurePosixPath(key)

    if self.file:
      self.template = Template(self, self.data, 'template')
      source = self.source

      if source is None:
        self.source = joinpath('.', dirname(owner.source))
      elif not source.startswith('./'):
        self.source = joinpath('.', self.source)

  def __call__(self, args:Mapping) -> list:
    name = self.format('name', args)

    if name is None:
      name = tojson(self.owner.properties.primarykey(args, sort_keys=True))
      name = 'mod-' + hashlib.new('md5', name.encode()).hexdigest()

    instance = Settings({'source': self.format('source', args)})

    for key in self.args:
      instance.update({key: get(args, key)})

    instance.merge(self.template(args))

    return ['module/' + name, instance]

  def _inherit(self):
    parent = self.owner.parent
    self.parent = parent.module if parent else parent
    file = self.file

    if file and parent:
      psource = Path(parent.source)

      if file.endswith('.tf.json'):
        self.file = psource.parent.joinpath(file)
      else:
        self.source = joinpath('.', psource.parent.joinpath(self.source))

  def format(self, key:str, args:Mapping):
    return self.owner.format(str(self.key / key), args)



from .exceptions import Error, PatternError
from .settings import get, getformat, merge, update, pop
from .template import jinja
from collections import UserDict
from collections.abc import Mapping
from json import dumps as json_encode, loads as json_decode
from os.path import dirname
from pathlib import PurePosixPath
from slugify import slugify as fnslugify
from zlib import adler32, crc32
import hashlib
import re

slugify_regex = r'[^-a-zA-Z0-9_]+'

def compute(prop:Mapping, value, args:Mapping):
  if value is not None:
    type_ = prop.get('type')

    if type_ in ['json', 'string'] and not isinstance(value, str):
      value = json_encode(value)

  translator = prop.get('translate')

  if translator:
    for v1, v2 in translator.items():
      if value == v1:
        value = v2
        break

  expr = prop.get('expr')

  if expr:
    try:
      value = jinja.compile_expression(expr)(this=value, **args)
    except Exception as e:
      raise Error('expr', *e.args)

  if not isinstance(value, str):
    return value

  patterns = prop.get('pattern')

  if patterns:
    if not isinstance(patterns, list):
      patterns = [patterns]

    for _ in patterns:
      _ = re.compile(_).match(value)

      if _:
        merge(args, _.groupdict({}))
        break

  algorithm = prop.get('hash')

  if algorithm:
    value = value.encode()
    fn = {"adler32": adler32, "crc32": crc32}.get(algorithm)
    value = '{:x}'.format(fn(value) & 0xffffffff) if fn else hashlib.new(algorithm, value).hexdigest()

  format_spec = prop.get('format')

  if format_spec:
    try:
      value = format_spec.format(value, **args)
    except KeyError as e:
      raise KeyError('format', *e.args)

  return value

def getvalue(prop:Mapping, args:Mapping, key:str):
  value = args.get(key)

  if value is None:
    value = getformat(prop, 'value', args)

  return value

def inherit(properties:Mapping):
  inherited = {}

  for key, prop in properties.items():
    if not prop.get('inherit', False):
      continue

    props = prop.get('properties')

    if props:
      props = inherit(props)

      if not props:
        continue

    prop = prop.copy()
    prop['ignore'] = True
    prop['sync'] = False

    if props:
      prop['properties'] = props

    prop.pop('alias', None)
    prop.pop('use', None)

    inherited[prop.get('alias', key)] = prop

  return inherited

def init(properties:Mapping, args:Mapping, defaults:bool=True, slugs:bool=True, root:Mapping=None) -> Mapping:
  if args is None:
    args = {}

  if root is None:
    root = args

  args_ = {'_': root, **args}

  for key, prop in properties.items():
    condition = prop.get('when')

    try:
      if condition and not jinja.compile_expression(condition)(**args_):
        continue
    except Exception as e:
      raise Error(key + '/when', *e.args)

    alias = prop.get('alias', key)

    try:
      value = getvalue(prop, args_, alias)
    except:
      value = None

    props = prop.get('properties')

    if props:
      try:
        value = merge(value, init(props, value, defaults, slugs, root), clone=False)
      except KeyError as e:
        raise KeyError(key, 'properties', *e.args)
      except Error as e:
        raise Error(key + '/properties/' + e.args[0], *e.args[1:])

      if not value:
        continue
    else:
      primary_key = prop.get('primary_key', False)

      if defaults and value is None:
        try:
          value = getformat(prop, 'default', args_)
        except:
          pass

      try:
        _ = compute(prop, value, args_)
      except KeyError as e:
        raise KeyError(key, *e.args)
      except Error as e:
        raise Error(key + '/' + e.args[0], *e.args[1:])

      if _ is None:
        if value is not None:
          pop(args, alias)
          pop(args_, alias)

        continue

      value = _

      if slugs and primary_key:
        alias_ = alias + '_'
        value_ = fnslugify(str(value), lowercase=False, regex_pattern=slugify_regex)
        args[alias_] = value_
        args_[alias_] = value_

    if isinstance(value, Mapping):
      unset = prop.get('unset', [])
      i = -1

      for item in unset:
        i += 1

        if isinstance(item, str):
          pop(value, item)
          continue

        condition = item.get('when')

        try:
          _ = (not condition or jinja.compile_expression(condition)(**args_))
        except Exception as e:
          raise Error(key + '/unset/' + i + '/when', *e.args)

        if _:
          pop(value, item.get('key'))

    args[alias] = value
    args_[alias] = value

  return args

def sync(properties:Mapping, settings:Mapping, root:Mapping=None) -> dict:
  if settings is None:
    settings = {}

  this = {}

  if root is None:
    root = this

  for key, prop in properties.items():
    sync_key = prop.get('sync', key)

    if sync_key is False:
      continue

    condition = prop.get('when')

    try:
      if condition and not jinja.compile_expression(condition)(_=root, **this):
        continue
    except Exception as e:
      raise Error(key + '/when', *e.args)

    if isinstance(sync_key, list):
      for k in sync_key:
        value = get(settings, k)

        if value is not None:
          break
    else:
      value = get(settings, sync_key)

    if value is None:
      continue

    alias = prop.get('alias', key)
    props = prop.get('properties')

    if props:
      value = sync(props, value, root)

      if not value:
        continue

    this[alias] = value

  return this

def tosettings(properties:Mapping, args:Mapping, defaults:bool=True, root:Mapping=None):
  if args is None:
    args = {}

  this = {}

  if root is None:
    root = args

  args_ = {'_': root, **args}

  for key, prop in properties.items():
    if prop.get('ignore', False):
      continue

    condition = prop.get('when')

    try:
      if condition and not jinja.compile_expression(condition)(**args_):
        continue
    except Exception as e:
      raise Error(key + '/when', *e.args)

    alias = prop.get('alias', key)
    value = args.get(alias)
    props = prop.get('properties')

    try:
      if props:
        try:
          value = tosettings(props, value, defaults, root)
        except Error as e:
          raise Error(key + '/properties/' + e.args[0], *e.args[1:])

        if not value:
          continue

        settings = value
      else:
        if value is None and defaults:
          try:
            value = getformat(prop, 'default', args_)
          except Exception as e:
            pass

          value = compute(prop, value, args_)

        type_ = prop.get('type', 'string')

        if type_ == 'list' or type_.startswith('list('):
          if value is not None:
            if isinstance(value, str):
              try:
                value = json_decode(value)
              except:
                pass

            if not isinstance(value, list):
              value = [value]

            if not value:
              value = None

          settings = value
        else:
          computed = prop.get('computed')

          if computed:
            try:
              settings = '${' + computed.format(value, **args_) + '}'
            except KeyError as e:
              raise KeyError('computed', *e.args)
          else:
            settings = value

        if settings is None:
          if defaults and prop.get('required', False):
            raise ValueError(alias)

          continue

      if value is not None:
        args[alias] = value
        args_[alias] = value

      key = prop.get('use', key)
      merge(this, update({}, {key: settings}), clone=False)
    except KeyError as e:
      raise KeyError(key, *e.args)

  return this

def walk(properties:Mapping, callback, prefix:str='') -> dict:
  for key, prop in properties.items():
    alias = prop.get('alias', key)
    props = prop.get('properties')

    if props:
      walk(props, callback, prefix + alias + '/')
    else:
      callback(prefix + alias, prop)

  return properties

class Properties(UserDict):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(get(cfg, key))
    self.owner = owner
    self.key = key
    self.context = PurePosixPath(owner.name, self.key)

    def inherit(alias, prop):
      if alias in owner.path:
        prop.setdefault('inherit', True)

    self.walk(inherit)

  def __call__(self, args:Mapping, defaults:bool=True, slugs:bool=True) -> dict:
    args = merge({}, args)

    try:
      return init(self, args, defaults=defaults, slugs=slugs)
    except KeyError as e:
      raise PatternError(self.owner.name, self.key, *e.args)
    except Error as e:
      raise Error(str(self.context / e.args[0]), *e.args[1:])

  def _inherit(self):
    resource = self.owner
    parent = resource.parent

    if parent:
      parent = parent.properties
      # Inherit missing path properties
      inherited = inherit(parent)

      if inherited:
        if resource.address == 'variable':
          for prop in inherited.values():
            prop.pop('computed', None)
            prop.setdefault('type', 'string')

            if prop.get('description'):
              prop.pop('ignore', False)

        self.data = merge(inherited, self.data)

    pattern = re.compile('\[([^[\]]+)\]')
    keys = re.findall('(?<!{){([^{}]+)}', resource.source)
    keys = [pattern.sub(r'/\1', _) for _ in keys]

    def setdefault(alias, prop):
      unset = []

      for key, value in prop.items():
        if value is None:
          unset.append(key)
      
      for key in unset:
        del prop[key]

      if alias in keys:
        prop.setdefault('inherit', True)
        prop.setdefault('primary_key', True)

    self.walk(setdefault)
    self.parent = parent

  def heritage(self, args:Mapping) -> dict:
    this = {}

    def callback(alias, prop):
      if prop.get('inherit', prop.get('primary_key', False)):
        value = get(args, alias)

        if value is not None:
          update(this, {alias: value})

    self.walk(callback)

    return this

  def primarykey(self, args:Mapping, slugs:Mapping=None, required:set=None) -> dict:
    this = {}

    def callback(alias, prop):
      if prop.get('primary_key', False):
        condition = prop.get('when')

        if condition:
          _ = dirname(alias)
          args_ = args if _ == '' else get(args, _)

          if not jinja.compile_expression(condition)(_=args, **args_):
            return

        value = get(args, alias)

        if value is None:
          if required is not None and prop.get('required', True):
            required.add(alias)
        else:
          update(this, {alias: value})

          if slugs is not None:
            alias += '_'
            value = get(args, alias)

            if value is not None:
              update(slugs, {alias: value})

    self.walk(callback)

    return this

  def remote(self, args:Mapping) -> dict:
    this = {}

    def callback(alias, prop):
      if prop.get('sync', True) != False:
        value = get(args, alias)

        if value is not None:
          update(this, {alias: value})        

    self.walk(callback)

    return this

  def sync(self, settings:Mapping) -> dict:
    return sync(self, settings)

  def tosettings(self, args:Mapping, defaults:bool=True) -> dict:
    try:
      settings = tosettings(self, args, defaults=defaults)
    except KeyError as e:
      raise PatternError(self.owner.name, self.key, *e.args)

    if self.owner.address == 'variable':
      for key, prop in self.items():
        if prop.get('ignore', False):
          continue

        key = prop.get('use', key)

        try:
          key.index('/')
          continue
        except ValueError:
          pass

        value = settings.get(key)

        if value is None and not prop.get('nullable', False):
          continue

        value = {'default': value}

        for k, _ in prop.items():
          if k in ['description', 'type', 'validation', 'sensitive']:
            value[k] = _

        settings[key] = value

    return settings.get('.', settings)

  def walk(self, callback):
    return walk(self, callback)



from collections import UserDict
from fnmatch import fnmatchcase
from io import StringIO
from json import dump as dump_json
from sys import stdout
from typing import Any, Sequence, Mapping
from yaml import dump as dump_yaml, load as load_yaml, FullLoader
import re

def format_map(format_spec:Any, args:Mapping) -> Any:
  if isinstance(format_spec, str):
    return format_spec.format_map(args)

  if isinstance(format_spec, Mapping):
    result = {}

    for key, value in format_spec.items():
      try:
        result[key] = format_map(value, args)
      except KeyError as e:
        raise KeyError(key, *e.args)

    return result

  if isinstance(format_spec, Sequence):
    result = []

    for i in range(len(format_spec)):
      try:
        result.append(format_map(format_spec[i], args))
      except KeyError as e:
        raise KeyError(str(i), *e.args)

    return result

  return format_spec

def get(settings:Any, path:str, default:Any = None, flatten:bool = True) -> Any:
  if isinstance(settings, Mapping):
    for key, value in settings.items():
      if key == path:
        return value

      if path.startswith(key + '/'):
        return get(value, path[len(key) + 1:], default, flatten)

    return default

  if isinstance(settings, Sequence) and not isinstance(settings, str):
    try:
      try:
        return settings[int(path)]
      except ValueError:
        pass

      try:
        i = path.index('/')
        return get(settings[int(path[0:i])], path[i + 1:], default, flatten)
      except ValueError:
        pass
    except IndexError:
      return default

    values = []

    for value in settings:
      value = get(value, path, default, flatten)

      if value is not None:
        if flatten and isinstance(value, Sequence) and not isinstance(value, str):
          values.extend(value)
        else:
          values.append(value)

    return values if values else default

  return default

def getformat(settings:Any, path:str, args:Mapping, default:Any = None) -> Any:
  try:
    value = format_map(get(settings, path), args)
    return default if value is None else value
  except KeyError as e:
    raise KeyError(path, *e.args)

def match(settings:Any, patterns:Any, literally:bool = False, default:bool = False) -> bool:
  if settings == patterns:
    return True

  if settings is None:
    return default

  if isinstance(patterns, Mapping):
    for key, pattern in patterns.items():
      if key == '$not':
        if match(settings, pattern, literally, default):
          return False
      else:
        if not isinstance(settings, Mapping):
          return False

        value = get(settings, key)

        if not match(value, pattern, literally, default):
          return False
    
    return True

  if isinstance(patterns, Sequence) and not isinstance(patterns, str):
    for pattern in patterns:
      if match(settings, pattern, literally, default):
        return True

    return False

  if literally:
    return False

  if isinstance(patterns, str) and fnmatchcase(str(settings), patterns):
    return True

  return False

def merge(result:Any, *others, extend:bool = False, clone:bool = True) -> Any:
  for other in others:
    if other is None:
      continue

    if isinstance(other, Mapping):
      if isinstance(result, Mapping):
        for key, value in other.items():
          if value is None:
            result[key] = value
          else:
            result[key] = merge(result.get(key), value, extend=extend, clone=clone)
      elif isinstance(result, Sequence) and not isinstance(result, str):
        try:
          for key, value in other.items():
            i = int(key)
            result[i] = merge(result[i], value, extend=extend, clone=clone)
        except:
          if clone:
            other = merge({}, other)

          if extend:
            result.append(other)
          else:
            result = other
      else:
        result = merge({}, other) if clone else other
    elif isinstance(other, Sequence) and not isinstance(other, str):
      if clone:
        other = [merge(None, _) for _ in other]

      if extend and isinstance(result, Sequence) and not isinstance(result, str):
        for value in other:
          if value not in result:
            result.append(value)
      else:
        result = other
    else:
      result = other

  return result

def pop(settings:Any, key:str, default = None, flatten:bool = True):
  if settings is None:
    return default

  try:
    if isinstance(settings, list):
      deleted = 0
      values = []

      for i in range(len(settings)):
        i -= deleted

        value = pop(settings[i], key, flatten=flatten)

        if not settings[i]:
          del settings[i]
          deleted += 1

        if value is None:
          continue

        if flatten and isinstance(value, list):
          values.extend(value)
        else:
          values.append(value)

      return values if values else default

    try:
      i = key.index('/')
      k = key[0:i]
      value = pop(settings[k], key[i + 1:], default, flatten)
      if not settings[k]: del settings[k]
      return value
    except ValueError:
      return settings.pop(key, default)
  except KeyError:
    return default

def update(settings:Any, other:Mapping) -> Mapping:
  if isinstance(settings, Sequence) and not isinstance(settings, str):
    try:
      for key, value in other.items():
        i = key.index('/')
        k = int(key[0:i])

        try:
          settings[k] = update(settings[k], {key[i + 1:]: value})
        except:
          settings[key] = value
    except:
      for item in settings:
        update(item, other)

    return settings

  if settings is None:
    settings = {}

  for key, value in other.items():
    try:
      i = key.index('/')
      k = key[0:i]
      settings[k] = update(settings.get(k), {key[i + 1:]: value})
    except:
      settings[key] = value

  return settings

def pprint(data, **opts):
  opts.setdefault('explicit_start', False)
  opts.setdefault('sort_keys', False)
  opts.setdefault('stream', stdout)
  dump_yaml(merge({}, data), **opts)

class Descriptor():
  def __init__(self, key, default=None):
    self.key = key
    self.default = default

  def __get__(self, obj, objtype=None):
    return obj.get(self.key, self.default)

  def __set__(self, obj, value):
    obj.update({self.key: self.default if value is None else value})

class Settings(UserDict):
  def __init__(self, data:Mapping=None, clone:bool=False, **opts):
    super().__init__()

    if clone:
      self.merge(data, clone=clone, **opts)
    elif data is not None:
      self.data = data

  def __str__(self):
    output = StringIO()
    self.print(stream=output)
    return output.getvalue()

  def copy(self):
    return Settings(self.data, extend=False, clone=True)

  def extend(self, *others, **opts):
    return self.merge(*others, extend=True, **opts)

  def filterkeys(self, pattern) -> Mapping:
    pattern = re.compile(pattern)
    return {k: v for k, v in self.items() if pattern.match(k)}

  def filter(self, patterns, true=True, literally=False):
    return {k: v for k, v in self.items() if match(v, patterns, true, literally)}

  def get(self, key, default=None, flatten=True):
    return get(self.data, key, default, flatten)

  def format(self, key, args, default=None):
    return getformat(self, key, args, default)

  def match(self, patterns, true=True, literally=False, default=None):
    return match(self, patterns, true=true, literally=literally, default=default)

  def merge(self, *others, **opts):
    merge(self.data, *others, **opts)
    return self

  def pop(self, key, default=None, flatten=True):
    return pop(self.data, key, default, flatten)

  def update(self, *others, **kwds):
    for other in others:
      update(self.data, other)

    update(self.data, kwds)

    return self

  def mapping(self, *keys):
    mapping = {}

    for k in keys:
      value = self.get(k)

      if value is not None:
        mapping[k] = value

    return mapping

  def print(self, **opts):
    pprint(self, **opts)

  @classmethod
  def dump(self, filename:str, data, **opts):
    data = merge({}, data)

    with open(filename, 'w') as fp:
      if filename.endswith('.json'):
        opts.setdefault('indent', 2)
        dump_json(data, fp, **opts)
        print(file=fp)
      else:
        dump_yaml(data, stream=fp, **opts)

  @classmethod
  def load(cls, filename:str, **opts):
    opts.setdefault('Loader', FullLoader)

    with open(filename) as fp:
      data = load_yaml(fp, **opts)

    return Settings(data)

class UserObject:
  pass



from . import Command
from ...settings import get, Descriptor
from collections.abc import Mapping
from json import loads as loadjson
from subprocess import check_output

class Show(Command):
  resources = Descriptor('resources')

  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key + '/show')

  def __call__(self, *args) -> Mapping:
    this = super().__call__('-json', '-no-color', *args)
    state = get(loadjson(check_output(this)), self.resources, [])
    return self.owner.state.update({self.cwd: state})

  def _inherit(self):
    super()._inherit()

    if self.resources is None:
      address = ['values/root_module']

      resource = self.owner
      parent = resource.parent

      while parent and resource.module and resource.module.file.endswith('.tf.json'):
        address.append('child_modules')
        resource = parent
        parent = resource.parent

      address.append('resources/address')
      self.resources = '/'.join(address)



from .exceptions import Error, PatternError
from .methods import Method, Methods
from .module import Module
from .path import VirtualPath
from .properties import Properties
from .settings import merge, pop, Descriptor, Settings
from .template import jinja, Template
from collections.abc import Mapping
from os import chdir
from os.path import dirname, join as joinpath
from pathlib import Path
import errno

class Resource(Settings):
  address = Descriptor('address')
  conflicts_with = Descriptor('conflicts_with')
  depends_on = Descriptor('depends_on')
  description = Descriptor('description')
  events = Descriptor('events')
  extends = Descriptor('extends')
  methods = Descriptor('methods')
  module = Descriptor('module')
  path = Descriptor('path')
  properties = Descriptor('properties')
  root = Descriptor('root')
  source = Descriptor('source')
  template = Descriptor('template')

  def __init__(self, owner, name:str, cfg:Mapping, parent=None):
    super().__init__()
    self.owner = owner
    self.parent = parent
    self.name = name

    depends_on = cfg.get('depends_on', [])

    if isinstance(depends_on, str):
      depends_on = [depends_on]

    self.depends_on = depends_on

    self.description = cfg.get('description', '')
    self.path = VirtualPath(self, cfg, 'path')
    self.root = ''
    self.source = cfg.get('source', 'locals.tf.json')
    self.address = cfg.get('address', 'locals')
    self.conflicts_with = cfg.get('conflicts_with')

    self.template = Template(self, cfg, 'template')
    self.properties = Properties(self, cfg, 'properties')
    self.methods = Methods(self, cfg, 'methods')
    self.events = cfg.get('events', {})
    self.module = Module(self, cfg, 'module')

    self._inherit()

  def __call__(self, command:str, *args, **kwds):
    return self.methods(command, *args, **kwds)

  def _inherit(self):
    parent = self.parent

    self.path._inherit()
    self.properties._inherit()
    self.methods._inherit()
    self.module._inherit()

    if parent:
      self.source = joinpath(dirname(parent.source), self.source)

    self.root = dirname(self.module.file if self.module.file and self.module.file.endswith('.tf.json') else self.source)
    state = None

    while parent and parent.root == self.root:
      state = parent.state
      parent = parent.parent

    if state is None:
      state = Settings()

    self.state = state

  def format(self, key, args, default=None):
    try:
      return super().format(key, args, default)
    except KeyError as e:
      raise PatternError(self.name, *e.args)
    except ValueError as e:
      raise Error(self.name, str(e))

  def beforesave(self, settings:Mapping):
    actions = self.events.get('onbeforesave', [])

    if not actions:
      return self

    i = 0

    if isinstance(actions, list):
      i_ = '/{}/'
    else:
      actions = [actions]
      i_ = '/'

    for action in actions:
      condition = action.get('when')

      try:
        if condition and not jinja.compile_expression(condition)(**settings):
          continue
      except Exception as e:
        raise Error(self.name + '/events/onbeforesave' + i_.format(i) + 'when', *e.args)

      keys = action.get('unset', [])

      if keys:
        if not isinstance(keys, list):
          keys = [keys]

        for key in keys:
          pop(settings, key)

    return self

  def trigger(self, event:str, args:Mapping, **options):
    event = 'on' + event
    props = self.properties

    for cmd_name, resources in self.events.get(event, {}).items():
      for resource_name, settings in resources.items():
        if not isinstance(settings, list):
          settings = [settings]

        i = -1

        for item in settings:
          i += 1
          condition = item.get('when')

          try:
            if condition and not jinja.compile_expression(condition)(**args):
              continue
          except Exception as e:
            raise Error(self.name + '/events/' + event + '/' + cmd_name + '/' + i + '/when', *e.args)

          resource = self.owner.load(resource_name)
          method = resource.methods.get(cmd_name)

          if not isinstance(method, Method):
            raise Error(self.name + '/events/' + event, 'No command named', cmd_name)

          if item.get('internal', resource_name.startswith('.')):
            _ = Settings(args, clone=True)
          else:
            _ = Settings(props.heritage(args)).merge(resource.properties.primarykey(args))

          unset = item.get('unset', [])

          if not isinstance(unset, list):
            unset = [unset]

          for key in unset:
            if isinstance(key, str):
              _.pop(key)
              continue

            condition = key.get('when')

            if not condition or jinja.compile_expression(condition)(**args):
              _.pop(key.get('key'))

          method(_.merge(item.get('args')), **merge({}, options, item.get('options', {})))

class Resources(dict):
  def __init__(self):
    super().__init__()

    root_dir = Path.cwd()
    config_dir = root_dir / ('.' + __package__)

    if not config_dir.is_dir():
      for root_dir in config_dir.parents:
        config_dir = root_dir / config_dir.stem

        if config_dir.is_dir():
          break

    self.paths = []
    name = config_dir.stem

    for parent in config_dir.parents:
      resources = parent / name / 'resources'

      if resources.is_dir():
        self.paths.append(resources)

    if not self.paths:
      raise FileNotFoundError(errno.ENOENT, 'Not a {} project (or any of the parent directories)'.format(__package__), name)

    chdir(root_dir)
    self.root_dir = root_dir

  def each(self, callback, parent=None, *args, **kwds):
    resources = []

    # Get all the the resources at the same level
    for resource in self.values():
      _ = resource.methods['sync'].parent
      _ = resource.parent if _ is None else _.owner

      if _ is parent:
        resources.append(resource)

    # Auxiliar variable to detect if it enters an infinite loop due to bad
    # `depends_on` configuration
    done = []
    count = 0
    skipped = 0

    while resources:
      execute = True
      resource = resources.pop(0)
      depends_on = resource.depends_on

      if depends_on:
        for dependency in depends_on:
          dependency = self.load(dependency)

          if dependency.name == resource.name:
            # Ignore if it depends on it self
            continue

          if dependency.name not in done:
            execute = False
            break

      if execute:
        done.append(resource.name)

        if resource.name.startswith('.'):
          count += self.each(lambda r: callback(r, *args, **kwds), resource)
        else:
          callback(resource, *args, **kwds)
          count += 1

        skipped = 0
      elif skipped < len(resources):
        # Puts the resource back on queue
        resources.append(resource)
        # Increases skipped by one in order to stop the cycle if unable to
        # resolve all dependencies
        skipped += 1
      else:
        # Raise error in case of bad configuration
        raise Error(resource.name + '/depends_on', str(depends_on))

    return count

  def extendConfig(self, cfg:Mapping, current=None):
    extends = cfg.pop('extends', None)

    if extends:
      if not isinstance(extends, list):
        extends = [extends]

      for name in extends:
        file, _ = self.loadConfig(name, current)
        cfg = self.extendConfig(merge(_, cfg, extend=True, clone=False), file)

    return cfg

  def load(self, name:str):
    if name is None:
      return None

    file = Path(name)

    if file.suffix in ['.yml', '.yaml']:
      name = file.stem

    resource = self.get(name)

    if resource is not None:
      return resource

    file, cfg = self.loadConfig(name)
    cfg = self.extendConfig(cfg, file)

    self[name] = resource = Resource(self, name, cfg, self.load(cfg.pop('parent', None)))

    return resource

  def loadAll(self):
    for file in self.paths[0].iterdir():
      if file.suffix in ['.yml', '.yaml'] and not file.name.startswith('.'):
        self.load(file.name)

    return self

  def loadConfig(self, name:str, current=None) -> list:
    for path in self.paths:
      file = path / name

      if file.suffix not in ['.yml', '.yaml']:
        file = file.with_suffix(file.suffix + '.yml')

        if not file.is_file():
          file = file.with_suffix('.yaml')

      if file.is_file() and file != current:
        return [file, Resource.load(file)]

    raise FileNotFoundError(errno.ENOENT, 'No such file', (self.paths[0] / file.with_suffix('.yml').stem).relative_to(self.root_dir).__str__())



from . import Method
from ..exceptions import Error, PatternError, RequiredArgument
from ..settings import match, merge, pprint, Descriptor
from ..template import jinja
from collections.abc import Mapping
from json import dumps as tojson
from shlex import split as splitcmd, join as joincmd
from subprocess import check_output
from yaml import load as loadyaml, FullLoader
from click import secho

class Sync(Method):
  when = Descriptor('when', {})

  class Chache:
    def __init__(self, cmd:str, settings:Mapping):
      self.cmd = cmd
      self.settings = settings

  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key + '/sync')
    self.cache = None

    for action in ['describe', 'list']:
      format_specs = self.data.get(action)

      if format_specs:
        if not isinstance(format_specs, list):
          format_specs = [format_specs]

        i = 0

        for format_spec in format_specs:
          if isinstance(format_spec, str):
            format_specs[i] = splitcmd(format_spec)
          i += 1

        self[action] = format_specs

  def _inherit(self):
    super()._inherit()

    parent = None
    resource = self.owner
    _ = resource.parent
    parent_name = self.get('parent')
    key = str(self.key)
    depends_on = None

    while _:
      parent = _.get(key)

      if parent and (parent_name is None or _.name == parent_name):
        break

      depends_on = _.name
      _ = _.parent

    if not parent and parent_name is not None:
      raise Error(str(self.context / 'parent'), 'No parent with the given name', parent_name)

    self.parent = parent

    if parent_name is not None and depends_on not in resource.depends_on:
      resource.depends_on.append(depends_on)

  def __call__(self, filters=None, **opts):
    terraform_import = opts.get('import', False)
    recursive = opts.get('recursive', False)
    resource = self.owner

    def update(args):
      args = resource('update', args, defaults=False, overwrite=True)

      if terraform_import:
        root = resource.format('root', args)
        state = resource.state.get(root)

        resource('terraform', 'chdir', root)

        if state is None:
          resource('terraform', 'init')
          resource('terraform', 'show')

        resource('terraform', 'import', args)
        resource.trigger('import', args)

      if recursive:
        args = resource.properties.heritage(args)
        resource.owner.each(lambda r, args: r('sync', args, **opts), resource, args)

    return self.describe(filters, update, force=opts.get('force', False))

  def describe(self, filters:Mapping, callback, **opts) -> int:
    return self.execute('describe', filters, callback, fallback='list', **opts)

  def list(self, filters:Mapping, callback, **opts) -> int:
    return self.execute('list', filters, callback, fallback='describe', **opts)

  def execute(self, action:str, filters:Mapping, callback, fallback:str=None, **opts) -> int:
    parent = self.parent
    format_specs = self.get(action)

    if not format_specs and fallback:
      format_specs = self.get(fallback)
      action = fallback
      fallback = None

    if not format_specs:
      return parent.list(filters, callback) if parent else 0

    i = 0
    count = len(format_specs)
    props = self.owner.properties
    args = props(filters, defaults=False, slugs=False)
    pprops = parent.owner.properties if parent else None
    context = str(self.context) + '.' + action + '()'

    while i < count:
      try:
        cmd_args = self.format('{}/{}'.format(action, i), args)
        err = None
        break
      except PatternError as e:
        print(e)
        err = e
        i += 1

    if err:
      count = 0

      if fallback and self.get(fallback):
        count = self.execute(fallback, filters, lambda _: self.execute(action, merge({}, filters, props.heritage(_)), callback))
      elif parent:
        count = parent.list(filters, lambda _: self.execute(action, merge({}, filters, pprops.heritage(_)), callback))

      if count > 0:
        return count

      pprint({context + '.arguments': args})
      raise err

    cmd = joincmd(cmd_args)

    if self.cache and self.cache.cmd == cmd:
      settings = self.cache.settings
      print('$', cmd, '(cached)')
    else:
      secho('$ {}'.format(cmd), bold=True)
      settings = loadyaml(check_output(cmd_args), Loader=FullLoader)
      self.cache = self.Chache(cmd, settings)

    if not settings:
      print(context + ': No objects found.')
      return 0

    if not isinstance(settings, list):
      settings = [settings]

    filters = props(filters, defaults=False, slugs=False)
    filters_ = props.remote(filters)
    heritage = pprops.heritage(filters) if self.parent else {}

    count = 0
    condition = self.when

    for args in settings:
      args = props.sync(args)
      args = merge(merge({}, heritage), args, clone=False)
      args_ = props(args, defaults=False, slugs=False)

      if parent:
        required = set()
        pprops.primarykey(args_, required=required)

        if required:
          required.clear()
          print(self.owner.name + ".primary_key:", props.primarykey(args_, required=required))
          print(self.owner.name + ": Missing argument:", ', '.join(required))
          _ = parent.list(pprops.heritage(args), lambda _: callback(merge(pprops.heritage(_), args, clone=False)))

          if _ > 0:
            count += _
            continue

          raise RequiredArgument(self.owner.name, required)

      if filters_:
        if not match(args_, filters_, literally=True):
          continue

      if condition and not opts.get('force', False):
        if isinstance(condition, str):
          try:
            if not jinja.compile_expression(condition)(**args):
              continue
          except Exception as e:
            raise Error(str(self.context / 'when'), *e.args)
        elif not match(args, condition):
          continue

      callback(args)
      count += 1

    if count == 0:
      ctx = str(self.context)

      if filters_:
        print('{}.{}(): No matches for the given filter: {}'.format(ctx, action, tojson(filters_)))
      else:
        print('{}/when: No matches for the given expression: {}'.format(ctx, condition))

    return count



from . import Command
from ...settings import pprint, Descriptor
from collections.abc import Mapping
from shlex import join as joincmd
from subprocess import check_call

class Import(Command):
  address = Descriptor('address')
  id = Descriptor('id')

  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key + '/import')

  def __call__(self, args:Mapping) -> str:
    if not self.id:
      return None

    context = str(self.context) + '()'

    try:
      address = self.format('address', args)
      id = self.format('id', args)
    except Exception as err:
      pprint({self.context: {'args': args}})
      raise err

    state = self.owner.state.get(self.cwd)

    if address in state:
      print(context + ':', 'Already managing a remote object for', address)
    else:
      this = super().__call__('-input=false', address, id)
      check_call(this)
      state.append(address)

    return id

  def __str__(self):
    this = super().args()
    this.extend(['-input=false', '{address}', '{id}'])
    return joincmd(this)

  def _inherit(self):
    super()._inherit()

    resource = self.owner
    address = self.address

    if address is None and resource.address.startswith('resource/'):
      address = '/'.join(resource.address.split('/')[1:3])
      address = address.replace('/', '.')

    if address:
      address = [address]

      parent = resource.parent

      while parent and resource.module and resource.module.file.endswith('.tf.json'):
        address.extend([resource.module.name, 'module'])
        resource = parent
        parent = resource.parent

      address = '.'.join(reversed(address))

    self.address = address



from ..exceptions import Error
from ..settings import get, Settings
from click import secho
from collections.abc import Mapping
from pathlib import PurePosixPath
from shlex import join as joincmd

class Method(Settings):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(get(cfg, key))
    self.owner = owner
    self.key = PurePosixPath(key)
    self.context = PurePosixPath(owner.name, self.key)

  def _inherit(self):
    parent = self.owner.parent

    if parent:
      parent = parent.get(str(self.key))

    self.parent = parent

  def format(self, key:str, args:Mapping):
    return self.owner.format(str(self.key / key), args)

class Group(Method):
  def __init__(self, owner, cfg:Mapping, key:str, methods:list):
    super().__init__(owner, cfg, key)

    for method in methods:
      method = method(owner, cfg, key)
      self[method.key.name] = method

  def __call__(self, command:str, *args, **kwds):
    method = self.get(command)

    if not isinstance(method, Method):
      raise Error(self.owner.name + '/' + str(self.key) + '()', 'No command named', command)

    return method(*args, **kwds)

  def _inherit(self):
    for method in self.values():
      method._inherit()

class Methods(Group):
  def __init__(self, owner, cfg:Mapping, key:str):
    from .sync import Sync
    from .terraform import Terraform
    from .update import Create, Update

    super().__init__(owner, cfg, key, [
      Create,
      Sync,
      Terraform,
      Update,
    ])

class ExternalCommand(Method):
  def __call__(self, *args) -> list:
    this = self.args(*args)
    secho('$ {}'.format(joincmd(this)), bold=True)
    return this

  def __str__(self):
    return joincmd(self.args())

  def args(self, *args) -> list:
    return args



class Error(Exception):
  """Generic tfadm error."""
  def __str__(self):
    return ': '.join(self.args)

class Required(Error):
  """Missing configuration key"""
  def __init__(self, context, *args):
    self.context = context
    super().__init__(*args)

  def __str__(self):
    return "{}: Missing configuration property: {}".format(self.context, '/'.join(self.args))

class RequiredArgument(Required):
  """Missing argument"""
  def __str__(self):
    return "{}: Missing argument: {}".format(self.context, '/'.join(self.args))

class PatternError(RequiredArgument):
  def __init__(self, context, *args):
    super().__init__(context, *(*args[:-1], '{' + args[-1] + '}'))



#!/usr/bin/env python

from sys import version_info, stdin

# TODO: Remove this check at some point in the future.
if version_info[0] < 3:
  raise ImportError('A recent version of Python 3 is required.')

from os.path import dirname, basename

__DIR__ = dirname(__file__)
__PACKAGE__ = basename(__DIR__)

if '__path__' not in globals():
  __path__ = [__DIR__]

from . import __version__
from .exceptions import Error, Required
from .resources import Resources, Resource
from .settings import merge
from collections.abc import Sequence
from subprocess import CalledProcessError
import click
import errno
import yaml

version_message = '{}, version {} from {} (Python {}.{})'.format(__PACKAGE__, __version__, __DIR__, *version_info[:2])

def main() -> int:
  try:
    cli()
    return 0
  except (Required, Error, yaml.scanner.ScannerError)  as e:
    code = 255
    click.secho('[Errno {}] {}'.format(code, str(e)), err=True, fg='red')
    return code
  except (FileNotFoundError, FileExistsError, PermissionError) as e:
    click.secho('[Errno {}] {}: {}'.format(*e.args, e.filename), err=True, fg='red')
    return e.errno
  except (KeyboardInterrupt, CalledProcessError):
    return errno.ECANCELED
  except SystemExit as code:
    return code

def read_args(resource:Resource, path:Sequence):
  paths = [path] if isinstance(path, str) else path

  if len(paths) == 1 and paths[0] == '-':
    return [_ for _ in yaml.safe_load_all(stdin)]

  args = []
  last = None
  path = None

  for path in paths:
    last = merge({}, last, clone=True)

    # Read arguments from standard input, if PATH is -
    if path == '-':
      # Overwrite PATH properties, if also set from standard input
      merge(last, yaml.safe_load(stdin), clone=False)
    else:
      # Properties linked to PATH are set automatically
      merge(last, resource.path(path), clone=False)
      args.append(last)

  if path == '-':
    args.append(last)

  return args

@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(__version__, '-V', '--version', prog_name=__PACKAGE__, message=version_message)
def cli():
  """Generates and modifies Terraform code in JSON format."""

@cli.command('help')
@click.pass_context
def cli_help(ctx):
  """Show this message and exit."""
  print(ctx.parent.get_help())

@cli.command('version')
def cli_version():
  """Show the version and exit."""
  print(version_message)

@cli.command('resources')
@click.argument('resource', required=False)
def cli_resources(resource=None):
  """Lists the available resources.

With RESOURCE, dump resource configuration to stdout.
"""
  resources = Resources()

  if resource is None:
    resources.loadAll()

    for name, resource in sorted(resources.items()):
      if not name.startswith('.'):
        print(name, resource.path)

        try:
          description = resource.description[0:resource.description.index("\n")]
        except:
          description = resource.description

        print(click.wrap_text(description, initial_indent='  ', subsequent_indent='  '), '\n')
  else:
    resource = _ = resources.load(resource)
    parents = []

    while _:
      parents.insert(0, _.name.replace('/', ':'))
      _ = _.parent

    print('---')
    print('#', '/'.join(parents))
    resource.print()

@cli.command('create')
@click.option(
  '--dry-run',
  default=False,
  help="Only print the object that would be saved, without saving it.",
  is_flag=True,
)
@click.option(
  '-o', '--overwrite',
  default=False,
  help="Overwrite the object, if exists.",
  is_flag=True,
)
@click.argument('resource')
@click.argument('path', required=False, nargs=-1)
def cli_create(resource, path=None, **opts):
  """Creates an object from stdin.

RESOURCE must be specified. Use 'tfadm resources' for a complete list of
available resources.

PATH is the virtual path of the the object. When PATH is -, reads object
properties from stdin.

If the object already exists, tfadm will error out, unless the '-o', or
'--overwrite' option is given.
"""
  resource = Resources().load(resource)

  for args in read_args(resource, path):
    resource('create', args, defaults=True, **opts)

@cli.command('update')
@click.option(
  '--dry-run',
  default=False,
  help="Only print the object that would be saved, without saving it.",
  is_flag=True,
)
@click.argument('resource')
@click.argument('path', required=False, nargs=-1)
@click.pass_context
def cli_update(ctx, resource, path=None, **opts):
  """Updates an object from stdin.

RESOURCE must be specified. Use 'tfadm resources' for a complete list of
available resources.

PATH is the filesystem path where the object is stored, relative to tfadm
project's root directory. When PATH is -, reads object attributes from stdin.

The object will be created if it doesn't exist.
"""
  resource = Resources().load(resource)

  for args in read_args(resource, path):
    resource('update', args, **opts)

@cli.command('sync')
@click.option(
  '-r', '--recursive',
  default=False,
  help='Copy resources and its sub-resources recursively.',
  is_flag=True,
)
@click.option(
  '--force',
  default=False,
  help='Ignores the `when` filter in the configuration, if any.',
  is_flag=True,
)
@click.option(
  '--import',
  default=False,
  help='Associate existing infrastructure with Terraform resources.',
  is_flag=True,
)
@click.argument('resource', required=False)
@click.argument('path', required=False, nargs=-1)
def cli_sync(resource, path=None, **opts):
  """Copies changes to the infrastructure into Terraform code.

Without RESOURCE, converts the existing infrastructure into Terraform code.

When RESOURCE or PATH is '-', filters existing infrastructure by stdin, copying
only matching objects. JSON and YAML formats are supported.

Use 'tfadm resources' for a complete list of available resources.
"""
  resources = Resources()

  if resource == '-' and not path:
    resource = None
    args = [_ for _ in yaml.safe_load_all(stdin)]
  elif resource is not None:
    resource = resources.load(resource)
    args = read_args(resource, path)

    if opts.get('recursive', False):
      resources.loadAll()
  else:
    args = None

  if not args:
    args = [None]

  for _ in args:
    if resource is None:
      resources.loadAll().each(lambda r, args: r('sync', args, **opts), None, _)
    else:
      resource('sync', _, **opts)

if __name__ == '__main__':
  exit(main())



from .settings import get, Settings, Descriptor
from collections.abc import Mapping
from jinja2 import Environment

jinja = Environment()

class Template(Settings):
  template = Descriptor('data')
  fields = Descriptor('fields')

  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(get(cfg, key))
    self.owner = owner
    self.key = key
    self.setdefault('data', {})
    self.setdefault('fields', [])

  def __call__(self, args:Mapping):
    values = {}

    for key in self.fields:
      values[key] = self.format('data/' + key, args)

    return Settings(self.template, clone=True, extend=False).update(values)

  def format(self, key:str, args:Mapping):
    return self.owner.format(self.key + '/' + key, args)



from . import Command
from collections.abc import Mapping
from subprocess import check_call

class Init(Command):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__(owner, cfg, key + '/init')

  def __call__(self, *args) -> bool:
    this = super().__call__('-input=false', *args)
    check_call(this)
    return self.owner.state.update({str(self.cwd): {}})



from .. import ExternalCommand, Group
from collections.abc import Mapping

class Command(ExternalCommand):
  def __call__(self, *args) -> list:
    return super().__call__(*args)

  def args(self, *args) -> list:
    this = ['terraform']

    if self.cwd:
      this.append('-chdir=' + self.cwd)

    this.append(self.key.name)
    this.extend(args)

    return this

  @classmethod
  def chdir(cls, path):
    cls.cwd = path
    return path

class Terraform(Group):
  def __init__(self, owner, cfg:Mapping, key:str):
    from .import_ import Import
    from .init import Init
    from .show import Show

    super().__init__(owner, cfg, key + '/terraform', [
      Import,
      Init,
      Show,
    ])

  def __call__(self, command:str, *args, **kwds):
    if command == 'chdir':
      return Command.chdir(*args, **kwds)

    return super().__call__(command, *args, **kwds)



from .settings import get, update
from collections import UserList
from collections.abc import Mapping
from parse import parse
from pathlib import Path, PurePosixPath
import re

class VirtualPath(UserList):
  def __init__(self, owner, cfg:Mapping, key:str):
    super().__init__()
    self.owner = owner
    self.key = key
    path = cfg.get(key, [])

    if isinstance(path, str):
      path = PurePosixPath(path).parts

    self.extend(path)

  def __call__(self, path:str) -> dict:
    this = self.__str__()
    resource = self.owner
    path = Path(path)

    if not resource.source.startswith(this) and path.is_dir():
      source = PurePosixPath(self.owner.source)
      mapping = parse(self.join(*source.parts[0:len(path.parts)]), self.join(*path.parts), case_sensitive=True)
      args = mapping.named if mapping else {}
      return args

    parts = path.parts

    mapping = {}
    last = len(self.data) - 1

    for i in range(len(parts)):
      if i >= last:
        mapping[self[i]] = self.join(*parts[i:])
        break
      mapping[self[i]] = parts[i]

    return update({}, mapping)

  def __str__(self) -> str:
    parts = []
    pattern = re.compile('/([^/]+)')

    for key in self.data:
      parts.append(pattern.sub(r'[\1]', key))

    return '{' + '}/{'.join(parts) + '}'

  def _inherit(self):
    parent = self.owner.parent

    if parent and parent.path:
      inherited = [*parent.path]

      def inherit(key, prop):
        if key in parent.path and not prop.get('inherit', True):
          inherited.remove(key)

      parent.properties.walk(inherit)

      if inherited:
        inherited.extend(self.data)
        self.data = inherited

    self.parent = parent

  def args(self, args:Mapping):
    this = {}

    for key in self.data:
      value = get(args, key)

      if value is not None:
        this[key] = value

    return this

  def extend(self, *others):
    for other in others:
      super().extend(other)
    return self

  def format_map(self, args):
    parts = []
    err = None

    for key in self.data:
      value = get(args, key)

      if value is None:
        err = KeyError(key)
      elif err is None:
        parts.append(value)
      else:
        raise err

    return self.join(*parts)

  @staticmethod
  def join(*parts:str):
    return '/'.join(parts)

  def parts(self, args:Mapping) -> list:
    parts = []

    for key in self.data:
      value = get(args, key)

      if value is None:
        break

      parts.append(value)

    return parts
